# ******************************************************************************
#  Copyright (c) 2021. Kneron Inc. All rights reserved.                        *
# ******************************************************************************

import numpy as np
import math

import ktc
EXEC_OPT = 1
'''
from utils.ExampleHelper import get_device_usb_speed_by_port_id
import kp
EXEC_OPT = 2
'''


MODEL_PATH = './'
MODEL_FILE = MODEL_PATH + 'models_720.nef'

DATA_PATH = './'
DATA_FILE = DATA_PATH + 'tri_filtData_151.npy'

BOARD_VID = 0x3231
BOARD_PID = 0x0720


def get_product_name(product_id: int) -> str:
    if product_id == kp.ProductId.KP_DEVICE_KL520.value:
        return '(KL520)'
    elif product_id == kp.ProductId.KP_DEVICE_KL720.value:
        return '(KL720)'
    elif product_id == 0x200:
        return '(KL720) (Please update firmware by Kneron DFUT)'
    else:
        return '(Unknown Device)'


def ktc_Inference():
    frame_data = np.load(DATA_FILE)
    print( 'fr_shape', frame_data.shape, 'type', type(frame_data[0][0][0][0]) )

    # check nef radix from quantization data
    radix = ktc.get_radix(frame_data)
    print( 'radix', radix )

    trigger_count = []
    pred_max = 0

    for idx in range( len(frame_data) ):
        input_data = [ frame_data[idx] ]
        '''
        '''
        print( np.array(input_data).shape, end='\t' )
        print( '%f %f %f %f' % (
                    input_data[0][0][0][0],
                    input_data[0][0][1][0],
                    input_data[0][0][2][0],
                    input_data[0][0][3][0]
                    ), end='\t'
                )

        # 5.2 E2E Simulator Check (Hardware)
        nef_predict = ktc.kneron_inference(
                input_data,
                nef_file = MODEL_FILE,
                platform = 720,
                radix = radix
                )[0][0][0]

        if nef_predict[0][1] > 0.5:
            print( idx, 'TRI\t', nef_predict[0][0], nef_predict[0][1] )
        else:
            print( idx, '\t', nef_predict[0][0], nef_predict[0][1] )


def KL720_Inference():
    usb_port_id = 0

    print('scanning kneron devices ...')
    device_list = kp.core.scan_devices()

    print('number of Kneron devices found: {}'.format(device_list.device_descriptor_number))

    if device_list.device_descriptor_number == 0:
        exit(0)

    print('listing devices infomation as follows:')
    for idx, device_descriptor in enumerate(device_list.device_descriptor_list):
        print()
        print('[{}] USB scan index: \'{}\''.format(idx, idx))
        print('[{}] USB port ID: \'{}\''.format(idx, device_descriptor.usb_port_id))
        print('[{}] Product ID: \'0x{:X} {}\''.format(idx,
                                                      device_descriptor.product_id,
                                                      get_product_name(product_id=device_descriptor.product_id)))

        print('[{}] USB link speed: \'{}\''.format(idx, device_descriptor.link_speed))
        print('[{}] USB port path: \'{}\''.format(idx, device_descriptor.usb_port_path))
        print('[{}] KN number: \'0x{:X}\''.format(idx, device_descriptor.kn_number))
        print('[{}] Connectable: \'{}\''.format(idx, device_descriptor.is_connectable))
        print('[{}] Firmware: \'{}\''.format(idx, device_descriptor.firmware))

        if device_descriptor.vendor_id == BOARD_VID and device_descriptor.product_id == BOARD_PID:
            usb_port_id = device_descriptor.usb_port_id
    print()

    """
    check device USB speed (Recommend run KL720 at super speed)
    """
    try:
        if kp.UsbSpeed.KP_USB_SPEED_SUPER != get_device_usb_speed_by_port_id(usb_port_id=usb_port_id):
            print('\033[91m' + '[Warning] Device is not run at super speed.' + '\033[0m')
    except Exception as exception:
        print('Error: check device USB speed fail, port ID = \'{}\', error msg: [{}]'.format(usb_port_id,
                                                                                             str(exception)))
        exit(0)

    """
    connect the device
    """
    try:
        print('[Connect Device]', 'w/ usb_port_id', usb_port_id)
        device_group = kp.core.connect_devices(usb_port_ids=[usb_port_id])
        print(' - Success')
    except kp.ApiKPException as exception:
        print('Error: connect device fail, port ID = \'{}\', error msg: [{}]'.format(usb_port_id,
                                                                                     str(exception)))
        exit(0)

    """
    setting timeout of the usb communication with the device
    """
    print('[Set Device Timeout]')
    kp.core.set_timeout(device_group=device_group, milliseconds=5000)
    print(' - Success')

    """
    prepare the image
    """
    print('[Read Image]')
    frame_data = np.load(DATA_FILE)
    #frame_data = np.einsum( 'ijkl->ikjl', frame_data )
    frame_num = len(frame_data)
    frame_size = len( bytes(frame_data[0].reshape(-1)) )
    print( 'fr_shape', frame_data.shape, 'fr_size', frame_size, 'type', type(frame_data[0][0][0][0]) )
    print(' - Success')

    """
    starting inference work
    """
    print('[Starting Inference Work]')
    print(' - Starting inference loop {} times'.format(frame_num))
    print(' - ', end='')
    inference_results = []
    for i in range(frame_num):
        data = frame_data[i].copy()

        # get model input radix, scale
        model_input_radix = model_nef_descriptor.models[0].input_nodes[0].quantization_parameters.quantized_fixed_point_descriptor_list[0].radix
        model_input_scale = model_nef_descriptor.models[0].input_nodes[0].quantization_parameters.quantized_fixed_point_descriptor_list[0].scale

        # get model input data layout
        model_input_data_layout = model_nef_descriptor.models[0].input_nodes[0].data_layout

        # get model input size (shape order: BxCxHxW)
        model_input_channel = model_nef_descriptor.models[0].input_nodes[0].shape_npu[1]
        model_input_height = model_nef_descriptor.models[0].input_nodes[0].shape_npu[2]
        model_input_width = model_nef_descriptor.models[0].input_nodes[0].shape_npu[3]

        # do normalization - please using your normalization process
        # for example: normalization: (data - 128) / 256.0
        # data = (data - 128) / 256.0

        # toolchain calculate the radix value from input data (after normalization), and set it into NEF model.
        # NPU will divide input data "2^radix" automatically, so, we have to scaling the input data here due to this reason.
        data *= (pow(2, model_input_radix) * model_input_scale)
        data = data.astype(np.int8)

        # re-layout the data to fit NPU data layout format
        # KL720 supported NPU input layout format: 4W4C8B, 1W16C8B, 16W1C8B
        if kp.ModelTensorDataLayout.KP_MODEL_TENSOR_DATA_LAYOUT_4W4C8B == model_input_data_layout:
            width_align_base = 4
            channel_align_base = 4
            pass
        elif kp.ModelTensorDataLayout.KP_MODEL_TENSOR_DATA_LAYOUT_1W16C8B == model_input_data_layout:
            width_align_base = 1
            channel_align_base = 16
            pass
        elif kp.ModelTensorDataLayout.KP_MODEL_TENSOR_DATA_LAYOUT_16W1C8B == model_input_data_layout:
            width_align_base = 16
            channel_align_base = 1
            pass
        else:
            print(' - Error: invalid input NPU layout format {}'.format(str(model_input_data_layout)))
            exit(0)

        # calculate width, channel alignment size
        model_input_width_align = width_align_base * math.ceil(model_input_width / float(width_align_base))
        model_input_channel_align = channel_align_base * math.ceil(model_input_channel / float(channel_align_base))

        # create re-layout data container
        re_layout_data = np.zeros((model_input_height, model_input_width_align, model_input_channel_align), dtype=np.int8)

        #print( data[0][0][0], data[0][1][0], data[0][2][0], data[0][3][0], end='\t' )
        # fill data in re-layout data container
        re_layout_data[:data.shape[0], :data.shape[1], :data.shape[2]] = data
        '''
        for idx_h in range( data.shape[0] ):
            for idx_w in range( data.shape[1] ):
                for idx_c in range( data.shape[2] ):
                    re_layout_data[idx_h][idx_w][idx_c*4] = data[idx_h][idx_w][idx_c]
        print( re_layout_data[0][0][0], re_layout_data[0][1][4], data[0][0][0], data[0][1][1], end='\t' )
        '''

        # convert re-layout data to npu inference buffer
        npu_input_buffer = re_layout_data.tobytes()

        # print(re_layout_data.shape, end='')

        """
        prepare app generic inference config
        """
        generic_raw_image_header = kp.v1.GenericRawBypassPreProcImageHeader(
            model_id=model_nef_descriptor.models[0].id,
            image_buffer_size=len(npu_input_buffer),
            inference_number=i
        )

        generic_raw_image_header.inference_number = i

        try:
            kp.v1.inference.generic_raw_inference_bypass_pre_proc_send(
                device_group=device_group,
                generic_raw_image_header=generic_raw_image_header,
                image_buffer=npu_input_buffer)

            generic_raw_bypass_pre_proc_result = kp.v1.inference.generic_raw_inference_bypass_pre_proc_receive(
                device_group=device_group,
                generic_raw_image_header=generic_raw_image_header,
                model_nef_descriptor=model_nef_descriptor)
            inference_results.append(generic_raw_bypass_pre_proc_result)
        except kp.ApiKPException as exception:
            print(' - Error: inference failed, error = {}'.format(exception))
            exit(0)

        print('.', end='', flush=True)
    print()

    """
    retrieve inference node output 
    """
    print( '[Retrieve Inference Node Output ]', '#{}'.format(generic_raw_bypass_pre_proc_result.header.num_output_node) )
    inf_per_frmae_node_output_list = []
    for inf_idx in range( len(inference_results) ):
        inference_float_node_output = kp.inference.generic_inference_retrieve_float_node(
            node_idx=0,
            generic_raw_result=inference_results[inf_idx],
            channels_ordering=kp.ChannelOrdering.KP_CHANNEL_ORDERING_CHW)
        inf_per_frmae_node_output_list.append(inference_float_node_output)

    print(' - Success')

    print('\r\n[Result]')
    '''
    print(inf_node_output_list)
    '''
    for out_idx in range( len(inf_per_frmae_node_output_list) ):
        '''
        '''
        print( '{:15f} {:15f} {:15f} {:15f}'.format(
                    frame_data[out_idx][0][0][0],
                    frame_data[out_idx][0][1][0],
                    frame_data[out_idx][0][2][0],
                    frame_data[out_idx][0][3][0]
                    ),
                end='\t'
                )

        out_data = inf_per_frmae_node_output_list[out_idx].ndarray
        #print( out_data.shape, out_data.dtype )
        # flat
        out_data = out_data.reshape(-1)
        if out_data[1] > 0.5:
            print( '{}\t{:15f} [ {:15f} ]\tTRI'.format( out_idx, out_data[0], out_data[1] ) )
        else:
            print( '{}\t{:15f}   {:15f}'.format( out_idx, out_data[0], out_data[1] ) )

    kp.core.disconnect_devices( device_group=device_group )


if __name__ == '__main__':
    if 1 == EXEC_OPT:
        ktc_Inference()
    elif 2 == EXEC_OPT:
        KL720_Inference()

